@@ -1,7 +1,8 @@
 """
 Functions and tools to unify types.
 
-__str__ of output types should be representation as type signature in function paramater
+Note that this uses dicts as ordered sets in many places so that input and output
+has the same ordering, so that git diffs are better
 """
 from __future__ import annotations
 
@@ -9,7 +10,8 @@ import typing
 import collections
 import pydantic
 import abc
-import ast
+
+import libcst as cst
 
 
 __all__ = [
@@ -33,15 +35,25 @@ __all__ = [
     "ClassMethodOutput",
     "UnionOutput",
     "BottomOutput",
+    "MethodDescriptorOutput",
+    "parser_config",
+    "NumpyUfuncOutput",
 ]
 
-# If thera are more than this amount in a union, just use any
+# If there are more than this amount in a union, just use any
 MAX_UNION_ITEMS = 5
+# If there are more than this amount in a string literal, just use any
+MAX_STRING_ITEMS = 5
 # More than this and it will be tuple of arbitrary length
 MAX_TUPLE_ITEMS = 2
 
 
-def annotation(tp: OutputType) -> typing.Optional[ast.AST]:
+parser_config = cst.PartialParserConfig(
+    python_version="3.8", encoding="utf-8", default_indent="    ", default_newline="\n",
+)
+
+
+def annotation(tp: OutputType) -> typing.Optional[cst.BaseExpression]:
     if is_unknown(tp):
         return None
     return tp.annotation
@@ -51,7 +63,7 @@ def create_type(o: object) -> OutputType:
     try:
         tp = pydantic.parse_obj_as(InputType, o)  # type: ignore
     except pydantic.error_wrappers.ValidationError:
-        raise ValueError(o)
+        raise ValueError(f"Could not parse JSON as input type: {o}")
     return to_output(tp)
 
 
@@ -60,32 +72,45 @@ def unify_inputs(types: typing.Iterable[InputType]) -> OutputType:
 
 
 def unify(types: typing.Iterable[OutputType]) -> OutputType:
-    types = set(types)
+    # groupby by the type to collect all types with same kind together
     tp_tps_to_tps: typing.DefaultDict[
-        typing.Type[OutputType], typing.Set[OutputType]
-    ] = collections.defaultdict(set)
-    for tp in types:
-        tp_tps_to_tps[type(tp)].add(tp)
+        # Use dict for values as ordered set to preserve ordering
+        typing.Type[OutputType],
+        typing.Dict[OutputType, None],
+    ] = collections.defaultdict(dict)
 
-    unified_types: typing.Set[OutputType] = set()
+    types = list(types)
+    while types:
+        tp = types.pop()
+        if isinstance(tp, ObjectOutput):
+            return ObjectOutput()
+        # If we have a union, add all to existing types
+        if isinstance(tp, UnionOutput):
+            types.extend(tp.options)
+            continue
+        tp_tps_to_tps[type(tp)][tp] = None
 
-    if ObjectOutput in tp_tps_to_tps:
-        return ObjectOutput()
+    # OK now we know that tp_tps_to_tps contains no union types
+    # Now try to unify each of the kinds to give us a set of types to add to to the unified
+    # Used a dict here as an ordered set
+    unified_types: typing.Dict[OutputType, None] = {}
 
+    # Sort so order of final types is preserved
     for tp_tp, tps in tp_tps_to_tps.items():
-        res: OutputType = _unify_output_tp(tp_tp, tps)
+        res: OutputType = _unify_output_tp(tp_tp, tps.keys())
         if isinstance(res, UnionOutput):
-            unified_types |= set(res.options)
+            for o in res.options:
+                unified_types[o] = None
         else:
-            unified_types.add(res)
+            unified_types[res] = None
 
     if not unified_types:
         return BottomOutput()
     if len(unified_types) > MAX_UNION_ITEMS:
         return ObjectOutput()
     if len(unified_types) == 1:
-        return unified_types.pop()
-    return UnionOutput(options=unified_types)
+        return next(iter(unified_types.keys()))
+    return UnionOutput(options=tuple(unified_types.keys()))
 
 
 OUTPUT_TYPE = typing.TypeVar("OUTPUT_TYPE", bound="OutputTypeBase")
@@ -117,9 +142,13 @@ class OutputTypeBase(BaseModel, abc.ABC):
 
     @property
     @abc.abstractmethod
-    def annotation(self) -> ast.AST:
+    def annotation(self) -> cst.BaseExpression:
         ...
 
+    @property
+    def module(self) -> typing.Optional[str]:
+        return None
+
 
 class InputTypeBase(BaseModel, abc.ABC):
     @abc.abstractmethod
@@ -146,7 +175,7 @@ class NoneOutput(OutputTypeBase):
 
     @property
     def annotation(self):
-        return ast.Constant(None, None)
+        return cst.Name("None")
 
 
 class StringInput(InputTypeBase):
@@ -157,37 +186,33 @@ class StringInput(InputTypeBase):
 
 
 class StringOutput(OutputTypeBase):
-    """
-    >>> str(StringOutput())
-    str
-    >>> str(StringOutput(options=["hi", "there"]))
-    Literal["hi", "there"]
-    """
 
     type: typing.Literal["str"] = "str"
-    options: typing.Union[typing.FrozenSet[str], None] = None
+    options: typing.Union[typing.Tuple[str, ...], None] = None
 
     @property
-    def annotation(self) -> ast.AST:
+    def annotation(self) -> cst.BaseExpression:
         if self.options:
-            return ast.Subscript(
-                ast.Name("Literal", ast.Load()),
-                ast.Index(
-                    ast.Tuple([ast.Constant(s, None) for s in self.options], ast.Load())
-                ),
-                ast.Load(),
+            return cst.Subscript(
+                cst.Name("Literal"),
+                [
+                    cst.SubscriptElement(cst.Index(cst.SimpleString(repr(option))))
+                    for option in self.options
+                ],
             )
-            return f'Literal[{", ".join(map(str, self.options))}]'
-        return ast.Name("int", ast.Load())
+        return cst.Name("str")
 
     @classmethod
     def unify(cls, tps: typing.Iterable[StringOutput]) -> StringOutput:
-        options: typing.Set[str] = set()
+        options: typing.Dict[str, None] = {}
         for tp in tps:
             if tp.options is None:
                 return StringOutput()
-            options.update(tp.options)
-        return StringOutput(options=options)
+            for o in tp.options:
+                options[o] = None
+            if len(options) > MAX_STRING_ITEMS:
+                return StringOutput()
+        return StringOutput(options=tuple(options.keys()))
 
 
 class ListInput(InputTypeBase):
@@ -202,11 +227,11 @@ class ListOutput(OutputTypeBase):
     item: OutputType
 
     @property
-    def annotation(self) -> ast.AST:
+    def annotation(self):
         if is_unknown(self.item):
-            return ast.Name("list", ast.Load())
-        return ast.Subscript(
-            ast.Name("List", ast.Load()), ast.Index(self.item.annotation), ast.Load()
+            return cst.Name("list")
+        return cst.helpers.parse_template_expression(
+            "List[{item}]", parser_config, item=self.item.annotation
         )
 
     @classmethod
@@ -228,21 +253,17 @@ class TupleOutput(OutputTypeBase):
     items: typing.Union[OutputType, typing.Tuple[OutputType, ...]]
 
     @property
-    def annotation(self) -> ast.AST:
+    def annotation(self):
         if is_unknown(self.items):
-            return ast.Name("tuple", ast.Load())
+            return cst.Name("tuple")
         if isinstance(self.items, tuple):
-            return ast.Subscript(
-                ast.Name("Tuple", ast.Load()),
-                ast.Index(ast.Tuple([s.annotation for s in self.items], ast.Load())),
-                ast.Load(),
+            return cst.Subscript(
+                cst.Name("Tuple"),
+                [cst.SubscriptElement(cst.Index(s.annotation)) for s in self.items],
             )
-        return ast.Subscript(
-            ast.Name("Tuple", ast.Load()),
-            ast.Index(
-                ast.Tuple([self.items.annotation, ast.Constant(..., None)], ast.Load())
-            ),
-            ast.Load(),
+
+        return cst.helpers.parse_template_expression(
+            "Tuple[{item}, ...]", parser_config, item=self.items.annotation
         )
 
     @classmethod
@@ -252,27 +273,25 @@ class TupleOutput(OutputTypeBase):
         length = None if None in lengths or len(lengths) != 1 else lengths.pop()
         if length is not None and length > MAX_UNION_ITEMS:
             length = None
-        # Is list if we have a fixed length typle
-        all_items: typing.Union[
-            typing.Set[OutputType], typing.List[typing.Set[OutputType]]
-        ] = [set() for _ in range(length)] if length is not None else set()
+        if length is None:
+            possible_values: typing.List[OutputType] = []
+            for tp in tps:
+                items = tp.items
+                if isinstance(items, tuple):
+                    possible_values.extend(items)
+                else:
+                    possible_values.append(items)
+            return TupleOutput(items=unify(possible_values))
+        i_to_possible_values: typing.List[typing.List[OutputType]] = [
+            [] for _ in range(length)
+        ]
         for tp in tps:
             items = tp.items
-            if isinstance(all_items, set):
-                if isinstance(items, tuple):
-                    all_items |= set(items)
-                else:
-                    all_items.add(items)
-            else:
-                assert isinstance(items, tuple)
-                for i, item in enumerate(items):
-                    all_items[i].add(item)
+            assert isinstance(items, tuple)
+            for i, item in enumerate(items):
+                i_to_possible_values[i].append(item)
 
-        return TupleOutput(
-            items=unify(all_items)
-            if isinstance(all_items, set)
-            else tuple(map(unify, all_items))
-        )
+        return TupleOutput(items=tuple(map(unify, i_to_possible_values)))
 
 
 class DictInput(InputTypeBase):
@@ -294,15 +313,15 @@ class DictOutput(OutputTypeBase):
     value: OutputType
 
     @property
-    def annotation(self) -> ast.AST:
+    def annotation(self):
         if is_unknown(self.key) and is_unknown(self.value):
-            return ast.Name("dict", ast.Load())
-        return ast.Subscript(
-            ast.Name("Dict", ast.Load()),
-            ast.Index(
-                ast.Tuple([self.key.annotation, self.value.annotation], ast.Load())
-            ),
-            ast.Load(),
+            return cst.Name("dict")
+        return cst.Subscript(
+            cst.Name("Dict"),
+            [
+                cst.SubscriptElement(cst.Index(self.key.annotation)),
+                cst.SubscriptElement(cst.Index(self.value.annotation)),
+            ],
         )
 
     @classmethod
@@ -316,8 +335,8 @@ class ObjectOutput(OutputTypeBase):
     type: typing.Literal["object"] = "object"
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Name("object", ast.Load())
+    def annotation(self):
+        return cst.Name("object")
 
     @classmethod
     def unify(cls, tps: typing.Iterable[ObjectOutput]) -> ObjectOutput:
@@ -339,7 +358,9 @@ NamedInput = typing.Union[BuiltinNamedInput, ModuleNamedInput]
 class OtherInputType(InputTypeBase):
     t: NamedInput
 
-    def to_output(self) -> typing.Union[OtherOutput, ObjectOutput]:
+    def to_output(self) -> typing.Union[OtherOutput, StringOutput, ObjectOutput]:
+        if self.t == BuiltinNamedInput(__root__="str"):
+            return StringOutput()
         return OtherOutput.safe_create(self.t)
 
 
@@ -352,15 +373,19 @@ class NamedOutput(BaseModel):
     def from_input(cls, input: NamedInput) -> typing.Optional[NamedOutput]:
         if isinstance(input, BuiltinNamedInput):
             return cls(name=input.__root__)
-        if '<' in input.name:
+        if "<" in input.name:
             return None
         return cls(name=input.name, module=input.module)
 
     @property
-    def annotation(self) -> ast.AST:
-        if self.module is None:
-            return ast.Name(self.name, ast.Load())
-        return ast.Attribute(ast.Name(self.module, ast.Load()), self.name, ast.Load())
+    def annotation(self) -> typing.Union[cst.Name, cst.Attribute]:
+        first_name, *rest = (
+            self.module.split(".") + [self.name] if self.module else [self.name]
+        )
+        expr: typing.Union[cst.Name, cst.Attribute] = cst.Name(first_name)
+        for name in rest:
+            expr = cst.Attribute(expr, cst.Name(name))
+        return expr
 
 
 class OtherOutput(OutputTypeBase):
@@ -372,18 +397,23 @@ class OtherOutput(OutputTypeBase):
         if tp is None:
             return ObjectOutput()
         return cls(type=tp)
+
     @property
-    def annotation(self) -> ast.AST:
+    def annotation(self):
         return self.type.annotation
 
     @classmethod
     def unify(
         cls, tps: typing.Iterable[OtherOutput]
     ) -> typing.Union[OtherOutput, UnionOutput]:
-        tps = set(tps)
+        tps = {t: None for t in tps}
         if len(tps) == 1:
-            return tps.pop()
-        return UnionOutput(options=tps)
+            return next(iter(tps.keys()))
+        return UnionOutput(options=tuple(tps.keys()))
+
+    @property
+    def module(self) -> typing.Optional[str]:
+        return self.type.module
 
 
 class ModuleInput(InputTypeBase):
@@ -399,16 +429,20 @@ class ModuleOutput(OutputTypeBase):
     name: typing.Optional[str] = None
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Attribute(ast.Name("types", ast.Load()), "ModuleType", ast.Load())
+    def annotation(self):
+        return cst.Attribute(cst.Name("types"), cst.Name("ModuleType"))
 
     @classmethod
     def unify(cls, tps: typing.Iterable[ModuleOutput]) -> ModuleOutput:
-        names = set(tp.name for tp in tps)
+        names = set(tps)
         if len(names) == 1:
-            return ModuleOutput(name=names.pop())
+            return names.pop()
         return ModuleOutput()
 
+    @property
+    def module(self) -> typing.Optional[str]:
+        return self.name
+
 
 class SliceInput(InputTypeBase):
     t: typing.Literal["slice"]
@@ -435,34 +469,32 @@ class SliceOutput(OutputTypeBase):
     step: OutputType
 
     @property
-    def annotation(self) -> ast.AST:
+    def annotation(self):
         """
         Doesn't exist yet as generic type, but should
 
         https://github.com/python/typing/issues/159
         """
         if is_unknown(self.start) and is_unknown(self.stop) and is_unknown(self.step):
-            return ast.Name("slice", ast.Load())
-        return ast.Subscript(
-            ast.Name("slice", ast.Load()),
-            ast.Index(
-                ast.Tuple(
-                    [self.start.annotation, self.stop.annotation, self.step.annotation],
-                    ast.Load(),
-                )
-            ),
-            ast.Load(),
+            return cst.Name("slice")
+        return cst.Subscript(
+            cst.Name("slice"),
+            [
+                cst.SubscriptElement(cst.Index(self.start.annotation)),
+                cst.SubscriptElement(cst.Index(self.stop.annotation)),
+                cst.SubscriptElement(cst.Index(self.start.annotation)),
+            ],
         )
 
     @classmethod
     def unify(cls, tps: typing.Iterable[SliceOutput]) -> SliceOutput:
-        start: typing.Set[OutputType] = set()
-        stop: typing.Set[OutputType] = set()
-        step: typing.Set[OutputType] = set()
+        start: typing.List[OutputType] = []
+        stop: typing.List[OutputType] = []
+        step: typing.List[OutputType] = []
         for tp in tps:
-            start.add(tp.start)
-            stop.add(tp.stop)
-            step.add(tp.step)
+            start.append(tp.start)
+            stop.append(tp.stop)
+            step.append(tp.step)
         return SliceOutput(start=unify(start), step=unify(step), stop=unify(stop))
 
 
@@ -480,11 +512,11 @@ class TypeOutput(OutputTypeBase):
     name: typing.Optional[NamedOutput] = None
 
     @property
-    def annotation(self) -> ast.AST:
+    def annotation(self):
         if self.name is None:
-            return ast.Name("type", ast.Load())
-        return ast.Subscript(
-            ast.Name("Type", ast.Load()), ast.Index(self.name.annotation), ast.Load()
+            return cst.Name("type")
+        return cst.Subscript(
+            cst.Name("Type"), [cst.SubscriptElement(cst.Index(self.name.annotation))]
         )
 
     @classmethod
@@ -494,6 +526,12 @@ class TypeOutput(OutputTypeBase):
             return TypeOutput(name=names.pop())
         return TypeOutput()
 
+    @property
+    def module(self) -> typing.Optional[str]:
+        if self.name:
+            return self.name.module
+        return None
+
 
 class FunctionInput(InputTypeBase):
     t: typing.Literal["function"]
@@ -519,8 +557,8 @@ class MethodWithoutSelfOutput(OutputTypeBase):
     name: str
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Name("Callable", ast.Load())
+    def annotation(self):
+        return cst.Name("Callable")
 
     @classmethod
     def unify(
@@ -531,6 +569,10 @@ class MethodWithoutSelfOutput(OutputTypeBase):
             return tps.pop()
         return FunctionOutput()
 
+    @property
+    def module(self) -> typing.Optional[str]:
+        return self.class_.name
+
 
 class FunctionOutput(OutputTypeBase):
     type: typing.Literal["function"] = "function"
@@ -538,8 +580,8 @@ class FunctionOutput(OutputTypeBase):
     name: typing.Optional[NamedOutput] = None
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Name("Callable", ast.Load())
+    def annotation(self):
+        return cst.Name("Callable")
 
     @classmethod
     def unify(cls, tps: typing.Iterable[FunctionOutput]) -> FunctionOutput:
@@ -548,6 +590,12 @@ class FunctionOutput(OutputTypeBase):
             return FunctionOutput(name=names.pop())
         return FunctionOutput()
 
+    @property
+    def module(self) -> typing.Optional[str]:
+        if self.name:
+            return self.name.module
+        return None
+
 
 class BuiltinFunctionInput(InputTypeBase):
     t: typing.Literal["builtin_function_or_method"] = "builtin_function_or_method"
@@ -576,8 +624,8 @@ class MethodOutput(OutputTypeBase):  # type: ignore
     self: OutputType
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Name("Callable", ast.Load())
+    def annotation(self):
+        return cst.Name("Callable")
 
     @classmethod
     def unify(
@@ -588,9 +636,13 @@ class MethodOutput(OutputTypeBase):  # type: ignore
             return tps.pop()
         return FunctionOutput()
 
+    @property
+    def module(self) -> typing.Optional[str]:
+        return self.self.module
+
 
 class MethodInput(InputTypeBase):
-    t: typing.Literal["method"] = "method"
+    t: typing.Literal["method"]
     v: MethodInputValue
 
     def to_output(self) -> MethodOutput:
@@ -598,13 +650,30 @@ class MethodInput(InputTypeBase):
 
 
 class MethodDescriptorInput(InputTypeBase):
-    t: typing.Literal["method_descriptor"] = "method_descriptor"
+    t: typing.Literal["method_descriptor"]
+    # TODO: Remove value, b/c its on signature
     v: MethodDescriptorInputValue
 
-    def to_output(self) -> ClassMethodOutput:
-        return ClassMethodOutput(
-            name=self.v.name, class_=self.v.class_.to_output().name
-        )
+    def to_output(self) -> MethodDescriptorOutput:
+        return MethodDescriptorOutput(name=self.v.name)
+
+
+class MethodDescriptorOutput(OutputTypeBase):  # type: ignore
+    type: typing.Literal["method_descriptor"] = "method_descriptor"
+    name: str
+
+    @property
+    def annotation(self):
+        return cst.Name("Callable")
+
+    @classmethod
+    def unify(
+        cls, tps: typing.Iterable[MethodDescriptorOutput]
+    ) -> typing.Union[MethodDescriptorOutput, FunctionOutput]:
+        tps = set(tps)
+        if len(tps) == 1:
+            return tps.pop()
+        return FunctionOutput()
 
 
 class MethodDescriptorInputValue(BaseModel):
@@ -618,8 +687,8 @@ class ClassMethodOutput(OutputTypeBase):  # type: ignore
     name: str
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Name("Callable", ast.Load())
+    def annotation(self):
+        return cst.Name("Callable")
 
     @classmethod
     def unify(
@@ -630,13 +699,33 @@ class ClassMethodOutput(OutputTypeBase):  # type: ignore
             return tps.pop()
         return FunctionOutput()
 
+    @property
+    def module(self) -> typing.Optional[str]:
+        return self.class_.module
+
 
 class NumpyUfuncInput(InputTypeBase):
     t: NumpyUfuncInputType
     v: str
 
-    def to_output(self) -> FunctionOutput:
-        return FunctionOutput(name=NamedOutput(module="numpy", name=self.v))
+    def to_output(self) -> NumpyUfuncOutput:
+        return NumpyUfuncOutput(name=self.v)
+
+
+class NumpyUfuncOutput(OutputTypeBase):
+    type: typing.Literal["numpy.ufunc"] = "numpy.ufunc"
+    name: typing.Optional[str] = None
+
+    @property
+    def annotation(self):
+        return NamedOutput(module="numpy", name="ufunc").annotation
+
+    @classmethod
+    def unify(cls, tps: typing.Iterable[NumpyUfuncOutput]) -> NumpyUfuncOutput:
+        names = set(tp.name for tp in tps)
+        if len(names) == 1:
+            return NumpyUfuncOutput(name=names.pop())
+        return NumpyUfuncOutput()
 
 
 class NumpyUfuncInputType(BaseModel):
@@ -679,24 +768,20 @@ class NumpyDTypeInput(InputTypeBase):
 
 class UnionOutput(OutputTypeBase):
     type: typing.Literal["union"] = "union"
+    # Can't be sets because when serializing, serialized as dicts
     options: typing.Tuple[OutputType, ...]
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Subscript(
-            ast.Name("Union", ast.Load()),
-            ast.Index(ast.Tuple([o.annotation for o in self.options], ast.Load())),
-            ast.Load(),
+    def annotation(self):
+        return cst.Subscript(
+            cst.Name("Union"),
+            [cst.SubscriptElement(cst.Index(o.annotation)) for o in self.options],
         )
 
     @classmethod
-    def unify(cls, tps: typing.Iterable[UnionOutput]) -> OutputType:
-        options: typing.Set[OutputType] = set()
-        for tp in tps:
-            options |= set(tp.options)
-        if len(options) == 1:
-            return options.pop()
-        return UnionOutput(options=options)
+    def unify(cls, unions: typing.Iterable[UnionOutput]) -> OutputType:
+        # This should never be called
+        raise NotImplementedError()
 
 
 class BottomOutput(OutputTypeBase):
@@ -712,8 +797,20 @@ class BottomOutput(OutputTypeBase):
         return UnionOutput(options=[])
 
     @property
-    def annotation(self) -> ast.AST:
-        return ast.Name("object", ast.Load())
+    def annotation(self):
+        return cst.Name("object")
+
+
+class OtherTypeInput(InputTypeBase):
+    """
+    Types that have different metaclasses will have a `t` that isn't `type` but their metaclass.
+    """
+
+    t: NamedInput
+    v: NamedInput
+
+    def to_output(self) -> TypeOutput:
+        return TypeOutput(name=NamedOutput.from_input(self.v))
 
 
 # Make them unions not subclasses so they are closed
@@ -737,6 +834,7 @@ InputType = typing.Union[
     NumpyConvert2MAInput,
     NumpyNDArrayInput,
     NumpyDTypeInput,
+    OtherTypeInput,
 ]
 OutputType = typing.Union[
     NoneOutput,
@@ -755,6 +853,8 @@ OutputType = typing.Union[
     ClassMethodOutput,
     UnionOutput,
     BottomOutput,
+    MethodDescriptorOutput,
+    NumpyUfuncOutput,
 ]
 
 for cls in (InputType.__args__ + OutputType.__args__) + (  # type: ignore
